@@ -193,6 +193,8 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
     char     **argp;
     char      *arg;
     char      *infile;
+    char       infile_abs[OSFNMAX];      /* fully-qualified input file name */
+    char       infile_path[OSFNMAX];         /* absolute path to input file */
     char      *exefile;            /* try with executable file if no infile */
     ulong      swapsize = 0xffffffffL;        /* allow unlimited swap space */
     int        swapena = OS_DEFAULT_SWAP_ENABLED;      /* swapping enabled? */
@@ -215,7 +217,7 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
     int        preload = FALSE;              /* TRUE => preload all objects */
     ulong      totsize;
     extern voccxdef *main_voc_ctx;
-    int        safety_level;                       /* file I/O safety level */
+    int        safety_read, safety_write;          /* file I/O safety level */
     char      *restore_file = 0;                    /* .SAV file to restore */
     char      *charmap = 0;                           /* character map file */
     int        charmap_none;       /* explicitly do not use a character set */
@@ -227,7 +229,7 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
     out_init();
 
     /* set safety level to 2 by default - read any/write current dir only */
-    safety_level = 2;
+    safety_read = safety_write = 2;
 
     /* no -ctab- yet */
     charmap_none = FALSE;
@@ -307,12 +309,20 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
                         trdusage_s(ec);
 
                     /* get the safety level from the argument */
-                    safety_level = atoi(p);
+                    safety_read = *p - '0';
+                    safety_write = (*(p+1) != '\0' ? *(p+1) - '0' :
+                                    safety_read);
+
+                    /* range-check the values */
+                    if (safety_read < 0 || safety_read > 4
+                        || safety_write < 0 || safety_write > 4)
+                        trdusage_s(ec);
 
                     /* tell the host system about the setting */
                     if (appctx != 0 && appctx->set_io_safety_level != 0)
                         (*appctx->set_io_safety_level)
-                            (appctx->io_safety_level_ctx, safety_level);
+                            (appctx->io_safety_level_ctx,
+                             safety_read, safety_write);
                 }
                 break;
                 
@@ -532,10 +542,14 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
     
     /* allocate stack and heap */
     totsize = (ulong)stksiz * (ulong)sizeof(runsdef);
-    if (totsize != (ushort)totsize)
+    if (totsize != (size_t)totsize)
         errsig1(ec, ERR_STKSIZE, ERRTINT, (uint)(65535/sizeof(runsdef)));
-    mystack = (runsdef *)mchalo(ec, (ushort)totsize, "runtime stack");
-    myheap = mchalo(ec, (ushort)heapsiz, "runtime heap");
+    mystack = (runsdef *)mchalo(ec, (size_t)totsize, "runtime stack");
+    myheap = mchalo(ec, heapsiz, "runtime heap");
+
+    /* get the absolute path for the input file */
+    os_get_abs_filename(infile_abs, sizeof(infile_abs), infile);
+    os_get_path_name(infile_path, sizeof(infile_path), infile_abs);
 
     /* set up execution context */
     runctx.runcxerr = ec;
@@ -557,6 +571,7 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
     runctx.runcxdmc = &supctx;
     runctx.runcxext = 0;
     runctx.runcxgamename = infile;
+    runctx.runcxgamepath = infile_path;
 
     /* set up setup context */
     supctx.supcxerr = ec;
@@ -585,7 +600,8 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
     bifctx.bifcxrnd = 0;
     bifctx.bifcxrndset = FALSE;
     bifctx.bifcxappctx = appctx;
-    bifctx.bifcxsafety = safety_level;
+    bifctx.bifcxsafetyr = safety_read;
+    bifctx.bifcxsafetyw = safety_write;
     bifctx.bifcxsavext = save_ext;
 
     /* initialize the regular expression parser context */
@@ -639,10 +655,32 @@ static void trdmain1(errcxdef *ec, int argc, char *argv[],
 
         /* vocctx is going out of scope - forget the global reference to it */
         main_voc_ctx = 0;
+
+        /* delete the voc context */
+        vocterm(&vocctx);
+
+        /* delete the undo context */
+        if (undoptr != 0)
+            objuterm(undoptr);
+
+        /* release the object cache structures */
+        mcmcterm(mctx);
+        mcmterm(globalctx);
     ERRENDCLN(ec)
 
     /* vocctx is going out of scope - forget the global reference to it */
     main_voc_ctx = 0;
+
+    /* delete the voc contxt */
+    vocterm(&vocctx);
+
+    /* delete the undo context */
+    if (undoptr != 0)
+        objuterm(undoptr);
+
+    /* release the object cache structures */
+    mcmcterm(mctx);
+    mcmterm(globalctx);
 }
 
 /*
