@@ -27,6 +27,8 @@ Modified
 #include "os.h"
 #include "t3std.h"
 #include "utf8.h"
+#include "osifcnet.h"
+
 
 /* ------------------------------------------------------------------------ */
 /*
@@ -182,38 +184,125 @@ const char *lib_find_vsn_suffix(const char *name_string, char suffix_char,
 
 
 /* ------------------------------------------------------------------------ */
+/*
+ *   allocating sprintf implementation 
+ */
+char *t3sprintf_alloc(const char *fmt, ...)
+{
+    /* package the arguments as a va_list and invoke our va_list version */
+    va_list args;
+    va_start(args, fmt);
+    char *str = t3vsprintf_alloc(fmt, args);
+    va_end(args);
+
+    /* return the allocated string */
+    return str;
+}
+
+/*
+ *   allocating vsprintf implementation 
+ */
+char *t3vsprintf_alloc(const char *fmt, va_list args)
+{
+    /* measure the required space - add in a byte for null termination */
+    size_t len = t3vsprintf(0, 0, fmt, args) + 1;
+
+    /* allocate space */
+    char *buf = (char *)t3malloc(len);
+    if (buf == 0)
+        return 0;
+
+    /* do the actual formatting */
+    t3vsprintf(buf, len, fmt, args);
+
+    /* return the allocated buffer */
+    return buf;
+}
+
 /*
  *   buffer-checked sprintf implementation
  */
-void t3sprintf(char *buf, size_t buflen, const char *fmt, ...)
+size_t t3sprintf(char *buf, size_t buflen, const char *fmt, ...)
 {
-    va_list args;
-
     /* package the arguments as a va_list and invoke our va_list version */
+    va_list args;
     va_start(args, fmt);
-    t3vsprintf(buf, buflen, fmt, args);
+    size_t len = t3vsprintf(buf, buflen, fmt, args);
     va_end(args);
+
+    /* return the length */
+    return len;
+}
+
+/* check for 'th' suffix in a format code */
+static const char *check_nth(const char *&fmt, int ival)
+{
+    /* presume no suffix */
+    const char *nth = "";
+
+    /* check for the 'th' suffix in the format code */
+    if (fmt[1] == 't' && fmt[2] == 'h')
+    {
+        /* skip the extra format characters */
+        fmt += 2;
+
+        /* 'th' suffix applies to most numbers */
+        nth = "th";
+
+        /* 
+         *   check for 1st, 2nd, 3rd, 21st, 22nd, 23rd, etc; but note that
+         *   the the teens are all th's 
+         */
+        if (ival % 100 < 10 || ival % 100 > 20)
+        {
+            switch (ival % 10)
+            {
+            case 1:
+                nth = "st";
+                break;
+                
+            case 2:
+                nth = "nd";
+                break;
+                
+            case 3:
+                nth = "rd";
+                break;
+            }
+        }
+    }
+
+    /* return the suffix we figured */
+    return nth;
 }
 
 /*
  *   buffer-checked vsprintf implementation 
  */
-void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
+size_t t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args0)
 {
     size_t rem;
+    size_t need = 0;
     char *dst;
 
-    /* if there's no room at all in the buffer, give up immediately */
-    if (buflen == 0)
-        return;
-    
+    /* 
+     *   make a private copy of the arguments, to ensure that we don't modify
+     *   the caller's copy (on some platforms, va_list is a reference type;
+     *   the caller might want to reuse their argument pointer for a two-pass
+     *   operation, such as a pre-format pass to measure how much space is
+     *   needed) 
+     */
+    va_list args;
+    os_va_copy(args, args0);
+
     /* scan the buffer */
-    for (dst = buf, rem = buflen - 1 ; *fmt != '\0' && rem != 0 ; ++fmt)
+    for (dst = buf, rem = buflen ; *fmt != '\0' ; ++fmt)
     {
         /* check for a format specifier */
         if (*fmt == '%')
         {
             const char *fmt_start = fmt;
+            const char *nth = "";
             const char *txt;
             size_t txtlen;
             char buf[20];
@@ -242,7 +331,7 @@ void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
                 ++fmt;
             }
 
-            /* if leading zeroes are desired, note it */
+            /* if leading zeros are desired, note it */
             if (*fmt == '0')
                 lead_char = '0';
 
@@ -361,9 +450,70 @@ void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
                  */
                 if (fld_wid != -1 && !left_align)
                 {
-                    for ( ; (size_t)fld_wid > txtlen && rem != 0 ;
-                          --fld_wid, --rem)
-                        *dst++ = ' ';
+                    for ( ; (size_t)fld_wid > txtlen ; --fld_wid)
+                    {
+                        ++need;
+                        if (rem > 1)
+                            --rem, *dst++ = ' ';
+                    }
+                }
+                break;
+
+            case 'P':
+                /* 
+                 *   URL parameter - this is a (char*) value with url
+                 *   encoding applied 
+                 */
+                txt = va_arg(args, char *);
+
+                /* if it's null, use an empty string */
+                if (txt == 0)
+                    txt = "";
+
+                /* use the field precision if given, or the string length */
+                txtlen = (fld_prec >= 0 ? fld_prec : strlen(txt));
+
+                /* encode the parameter and add it to the buffer */
+                for ( ; txtlen != 0 ; --txtlen, ++txt)
+                {
+                    switch (*txt)
+                    {
+                    case '!':
+                    case '*':
+                    case '\'':
+                    case '(':
+                    case ')':
+                    case ';':
+                    case ':':
+                    case '@':
+                    case '&':
+                    case '=':
+                    case '+':
+                    case '$':
+                    case ',':
+                    case '/':
+                    case '?':
+                    case '#':
+                    case '[':
+                    case ']':
+                    case ' ':
+                        /* use % encoding for special characters */
+                        sprintf(buf, "%%%02x", (unsigned)(uchar)*txt);
+                        need += 3;
+                        for (i = 0 ; i < 3 ; ++i)
+                        {
+                            if (rem > 1)
+                                *dst++ = buf[i], --rem;
+                        }
+                        break;
+
+                    default:
+                        /* copy anything else as-is */
+                        need += 1;
+                        if (rem > 1)
+                            *dst++ = *txt, --rem;
+                        break;
+                    }
                 }
                 break;
 
@@ -376,7 +526,14 @@ void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
 
             case 'd':
                 /* the value is an int, formatted in decimal */
-                sprintf(buf, "%d", va_arg(args, int));
+                {
+                    int ival = va_arg(args, int);
+                    sprintf(buf, "%d", ival);
+
+                    /* check for '%dth' notation (1st, 2nd, etc) */
+                    nth = check_nth(fmt, ival);
+                }
+                /* fall through to num_common: */
                 
             num_common:
                 /* use the temporary buffer where we formatted the value */
@@ -384,32 +541,38 @@ void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
                 txtlen = strlen(buf);
 
                 /* 
-                 *   Pad with leading spaces or zeroes if the requested
+                 *   Pad with leading spaces or zeros if the requested
                  *   field width exceeds the actual size and we have the
                  *   default left alignment.  
                  */
                 if (fld_wid != -1 && !left_align && (size_t)fld_wid > txtlen)
                 {
                     /* 
-                     *   if we're showing leading zeroes, and we have a
+                     *   if we're showing leading zeros, and we have a
                      *   negative number, show the '-' first 
                      */
-                    if (lead_char == '0' && txt[0] == '-' && rem != 0)
+                    if (lead_char == '0' && txt[0] == '-')
                     {
                         /* add the '-' at the start of the output */
-                        *dst++ = '-';
-                        --rem;
-
+                        ++need;
+                        if (rem > 1)
+                            --rem, *dst++ = '-';
+                        
                         /* we've shown the '-', so skip it in the number */
                         ++txt;
                         --txtlen;
                     }
-
+                    
                     /* add the padding */
-                    for ( ; (size_t)fld_wid > txtlen && rem != 0 ;
-                          --fld_wid, --rem)
-                        *dst++ = lead_char;
+                    for ( ; (size_t)fld_wid > txtlen ; --fld_wid)
+                    {
+                        ++need;
+                        if (rem > 1)
+                            --rem, *dst++ = lead_char;
+                    }
                 }
+
+                /* done */
                 break;
 
             case 'u':
@@ -469,15 +632,31 @@ void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
             }
 
             /* add the text to the buffer */
-            for (i = txtlen ; i != 0 && rem != 0 ; --i, --rem)
-                *dst++ = *txt++;
+            for (i = txtlen ; i != 0 ; --i, ++txt)
+            {
+                ++need;
+                if (rem > 1)
+                    --rem, *dst++ = *txt;
+            }
+
+            /* add the 'nth' suffix, if applicable */
+            for ( ; *nth != 0 ; ++nth)
+            {
+                ++need;
+                if (rem > 1)
+                    --rem, *dst++ = *nth;
+            }
 
             /* add an ellipsis if desired */
             if (add_ellipsis)
             {
                 /* add three '.' characters */
-                for (i = 3 ; i != 0 && rem != 0 ; --i, --rem)
-                    *dst++ = '.';
+                for (i = 3 ; i != 0 ; --i)
+                {
+                    ++need;
+                    if (rem > 1)
+                        --rem, *dst++ = '.';
+                }
 
                 /* for padding purposes, count the ellipsis */
                 txtlen += 3;
@@ -490,23 +669,45 @@ void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
             if (left_align && fld_wid != -1 && (size_t)fld_wid > txtlen)
             {
                 /* add spaces to pad out to the field width */
-                for (i = fld_wid ; i > txtlen && rem != 0 ; --i, --rem)
-                    *dst++ = ' ';
+                for (i = fld_wid ; i > txtlen ; --i)
+                {
+                    ++need;
+                    if (rem > 1)
+                        --rem, *dst++ = ' ';
+                }
             }
         }
         else
         {
             /* it's not a format specifier - just copy it literally */
-            *dst++ = *fmt;
-            --rem;
+            ++need;
+            if (rem > 1)
+                --rem, *dst++ = *fmt;
         }
     }
 
     /* add the trailing null */
-    *dst = '\0';
+    if (rem != 0)
+        *dst = '\0';
+
+    /* bracket the va_copy at the top of the function */
+    os_va_copy_end(args);
+
+    /* return the needed length */
+    return need;
 }
 
 
+/* ------------------------------------------------------------------------ */
+/*
+ *   Convert string to lower case 
+ */
+void t3strlwr(char *p)
+{
+    for ( ; *p != '\0' ; ++p)
+        *p = (char)to_lower(*p);
+}
+
 /* ------------------------------------------------------------------------ */
 /*
  *   Debugging routines for memory management 
@@ -517,6 +718,18 @@ void t3vsprintf(char *buf, size_t buflen, const char *fmt, va_list args)
 #include <stdio.h>
 #include <stdlib.h>
 
+#define T3_DEBUG_MEMGUARD
+#ifdef T3_DEBUG_MEMGUARD
+#define MEM_GUARD_PREFIX  char guard[sizeof(mem_prefix_t *)];
+#define MEM_GUARD_POST_BYTES   sizeof(mem_prefix_t *)
+#else /* T3_DEBUG_MEMGUARD */
+#define MEM_GUARD_PREFIX
+#define MEM_GUARD_BYTES   0
+#define mem_check_guard(blk)
+#define mem_set_guard(blk)
+#endif /* T3_DEBUG_MEMGUARD */
+
+
 /*
  *   memory block prefix - each block we allocate has this prefix attached
  *   just before the pointer that we return to the program 
@@ -527,12 +740,51 @@ struct mem_prefix_t
     size_t siz;
     mem_prefix_t *nxt;
     mem_prefix_t *prv;
+    int alloc_type;
+    OS_MEM_PREFIX
+    MEM_GUARD_PREFIX
 };
 
+#ifdef T3_DEBUG_MEMGUARD
+static void mem_make_guard(char *b, const mem_prefix_t *blk)
+{
+    memcpy(b, &blk, sizeof(blk));
+    const static unsigned char x[] =
+        { 0xcf, 0xfd, 0xef, 0xfb, 0xbf, 0xfc, 0xdf, 0xfe  };
+    for (size_t i = 0 ; i < sizeof(blk) ; ++i)
+        b[i] ^= x[i % countof(x)];
+}
+static void mem_set_guard(mem_prefix_t *blk)
+{
+    char b[sizeof(mem_prefix_t *)];
+    mem_make_guard(b, blk);
+    memcpy(blk->guard, b, sizeof(b));
+    memcpy((char *)(blk + 1) + blk->siz, b, sizeof(b));
+}
+static void mem_check_guard(const mem_prefix_t *blk)
+{
+    char b[sizeof(mem_prefix_t *)];
+    mem_make_guard(b, blk);
+    if (memcmp(blk->guard, b, sizeof(b)) != 0)
+        fprintf(stderr, "pre guard bytes corrupted: addr=%lx, id=%ld, siz=%lu "
+                OS_MEM_PREFIX_FMT "\n",
+                (long)(blk + 1), blk->id, (unsigned long)blk->siz
+                OS_MEM_PREFIX_FMT_VARS(blk));
+    if (memcmp((char *)(blk + 1) + blk->siz, b, sizeof(b)) != 0)
+        fprintf(stderr, "post guard bytes corrupted: addr=%lx, id=%ld, siz=%lu "
+                OS_MEM_PREFIX_FMT "\n",
+                (long)(blk + 1), blk->id, (unsigned long)blk->siz
+                OS_MEM_PREFIX_FMT_VARS(blk));
+}
+#endif /* T3_DEBUG_MEMGUARD */
+
 /* head and tail of memory allocation linked list */
 static mem_prefix_t *mem_head = 0;
 static mem_prefix_t *mem_tail = 0;
 
+/* mutex for protecting the memory tracker list */
+static OS_Mutex mem_mutex;
+
 /*
  *   Check the integrity of the heap: traverse the entire list, and make
  *   sure the forward and backward pointers match up.  
@@ -541,6 +793,9 @@ static void t3_check_heap()
 {
     mem_prefix_t *p;
 
+    /* lock the memory mutex while accessing the memory block list */
+    mem_mutex.lock();
+
     /* scan from the front */
     for (p = mem_head ; p != 0 ; p = p->nxt)
     {
@@ -555,29 +810,41 @@ static void t3_check_heap()
             || (p->nxt == 0 && p != mem_tail))
             fprintf(stderr, "\n--- heap corrupted ---\n");
     }
+
+    /* done with the list */
+    mem_mutex.unlock();
 }
 
 /*
  *   Allocate a block, storing it in a doubly-linked list of blocks and
  *   giving the block a unique ID.  
  */
-void *t3malloc(size_t siz)
+void *t3malloc(size_t siz, int alloc_type)
 {
     static long id;
     static int check = 0;
     mem_prefix_t *mem;
 
     /* allocate the memory, including its prefix */
-    mem = (mem_prefix_t *)malloc(siz + sizeof(mem_prefix_t));
+    mem = (mem_prefix_t *)malloc(siz + sizeof(mem_prefix_t)
+                                 + MEM_GUARD_POST_BYTES);
 
     /* if that failed, return failure */
     if (mem == 0)
         return 0;
 
+    /* set the guard bytes */
+    mem->alloc_type = alloc_type;
+    mem->siz = siz;
+    mem_set_guard(mem);
+
+    /* lock the memory mutex while accessing the memory block list */
+    mem_mutex.lock();
+
     /* set up the prefix */
     mem->id = id++;
-    mem->siz = siz;
     mem->prv = mem_tail;
+    os_mem_prefix_set(mem);
     mem->nxt = 0;
     if (mem_tail != 0)
         mem_tail->nxt = mem;
@@ -585,6 +852,9 @@ void *t3malloc(size_t siz)
         mem_head = mem;
     mem_tail = mem;
 
+    /* done with the list */
+    mem_mutex.unlock();
+
     /* check the heap for corruption if desired */
     if (check)
         t3_check_heap();
@@ -604,7 +874,7 @@ void *t3realloc(void *oldptr, size_t newsiz)
     size_t oldsiz;
 
     /* allocate a new block */
-    newptr = t3malloc(newsiz);
+    newptr = t3malloc(newsiz, T3MALLOC_TYPE_MALLOC);
 
     /* copy the old block into the new block */
     oldsiz = (((mem_prefix_t *)oldptr) - 1)->siz;
@@ -619,26 +889,50 @@ void *t3realloc(void *oldptr, size_t newsiz)
 
 
 /* free a block, removing it from the allocation block list */
-void t3free(void *ptr)
+void t3free(void *ptr, int alloc_type)
 {
+    /* statics for debugging */
     static int check = 0;
     static int double_check = 0;
     static int check_heap = 0;
-    mem_prefix_t *mem = ((mem_prefix_t *)ptr) - 1;
     static long ckblk[] = { 0xD9D9D9D9, 0xD9D9D9D9, 0xD9D9D9D9 };
-    size_t siz;
+
+    /* ignore freeing null */
+    if (ptr == 0)
+        return;
 
     /* check the integrity of the entire heap if desired */
     if (check_heap)
         t3_check_heap();
 
+    /* get the prefix */
+    mem_prefix_t *mem = ((mem_prefix_t *)ptr) - 1;
+    size_t siz;
+
+    /* 
+     *   check that the call type matches the allocating call type (malloc,
+     *   new, or new[]) 
+     */
+    if (mem->alloc_type != alloc_type)
+        fprintf(stderr, "\n--- memory block freed with wrong call type: "
+                "block=%lx, size=%d, id=%lu, alloc type=%d, free type=%d ---\n",
+                (unsigned long)ptr, mem->siz, mem->id,
+                mem->alloc_type, alloc_type);
+
     /* check for a pre-freed block */
     if (memcmp(mem, ckblk, sizeof(ckblk)) == 0)
     {
-        fprintf(stderr, "\n--- memory block freed twice ---\n");
+        fprintf(stderr, "\n--- memory block freed twice: %lx ---\n",
+                (unsigned long)ptr);
         return;
     }
 
+    /* check the guard bytes for overwrites */
+    mem_check_guard(mem);
+
+    /* lock the memory mutex while accessing the memory block list */
+    mem_mutex.lock();
+
     /* if desired, check to make sure the block is in our list */
     if (check)
     {
@@ -649,7 +943,9 @@ void t3free(void *ptr)
                 break;
         }
         if (p == 0)
-            fprintf(stderr, "\n--- memory block not found in t3free ---\n");
+            fprintf(stderr,
+                    "\n--- memory block not found in t3free: %lx ---\n",
+                    (unsigned long)ptr);
     }
 
     /* unlink the block from the list */
@@ -680,6 +976,9 @@ void t3free(void *ptr)
                     "t3free ---\n");
     }
 
+    /* done with the list */
+    mem_mutex.unlock();
+
     /* make it obvious that the memory is invalid */
     siz = mem->siz;
     memset(mem, 0xD9, siz + sizeof(mem_prefix_t));
@@ -716,16 +1015,54 @@ void t3_list_memory_blocks(void (*cb)(const char *))
     /* display introductory message */
     (*cb)("\n(T3VM) Memory blocks still in use:\n");
 
+    /* lock the memory mutex while accessing the memory block list */
+    mem_mutex.lock();
+
     /* display the list of undeleted memory blocks */
     for (mem = mem_head, cnt = 0 ; mem ; mem = mem->nxt, ++cnt)
     {
-        sprintf(buf, "  id = %ld, siz = %zu\n", mem->id, mem->siz);
+        sprintf(buf, "  addr=%lx, id=%ld, siz=%lu" OS_MEM_PREFIX_FMT "\n",
+                (long)(mem + 1), mem->id, (unsigned long)mem->siz
+                OS_MEM_PREFIX_FMT_VARS(mem));
         (*cb)(buf);
     }
 
+    /* done with the mutex */
+    mem_mutex.unlock();
+
     /* display totals */
     sprintf(buf, "\nTotal blocks in use: %d\n", cnt);
     (*cb)(buf);
 }
 
+
+#ifdef __WIN32__
+/*
+ *   Windows-specific additions to the memory header.  We'll track the first
+ *   couple of return addresses from the stack, to make it easier to track
+ *   down where the allocation request came from.  
+ */
+
+void os_mem_prefix_set(mem_prefix_t *mem)
+{
+    /* 
+     *   Trace back the call stack.  In the standard Intel stack arrangement,
+     *   BP is the base pointer for the frame, and points to the enclosing
+     *   frame pointer.  Just above BP is the return address.  We're not
+     *   interested in our own return address, so skip the first frame.  
+     */
+    DWORD bp_;
+    __asm mov bp_, ebp;
+    bp_ = IsBadReadPtr((DWORD *)bp_, sizeof(bp_)) ? 0 : *(DWORD *)bp_;
+
+    for (size_t i = 0 ;
+         i < countof(mem->stk) && !IsBadReadPtr((DWORD *)bp_, sizeof(bp_))
+             && bp_ < *(DWORD *)bp_ ;
+         ++i, bp_ = *(DWORD *)bp_)
+        mem->stk[i].return_addr = ((DWORD *)bp_)[1];
+}
+
+#endif /* __WIN32__ */
+
+
 #endif /* T3_DEBUG */
